@@ -1,12 +1,22 @@
 from copy import deepcopy
+from shutil import copyfileobj
+import numpy as np
+from scipy.interpolate import interp1d
+from astropy.io import fits
+from astropy.wcs import WCS
+from astropy import units as u
+
+from synphot.units import PHOTLAM
 
-from .. import rc
-from ..commands.user_commands import UserCommands
 from .optics_manager import OpticsManager
 from .fov_manager import FOVManager
 from .image_plane import ImagePlane
+from ..commands.user_commands import UserCommands
 from ..detector import DetectorArray
+from ..source.source import Source
 from ..utils import from_currsys
+from .. import rc
+from . import fov_utils as fu
 
 
 class OpticalTrain:
@@ -63,7 +73,6 @@ class OpticalTrain:
         >>> opt["dark_current"] = {"value": 0.75, "dit": 30}
 
     """
-
     def __init__(self, cmds=None):
 
         self._description = self.__repr__()
@@ -75,7 +84,6 @@ class OpticalTrain:
         self.yaml_dicts = None
         self._last_source = None
 
-
         if cmds is not None:
             self.load(cmds)
 
@@ -121,6 +129,7 @@ class OpticalTrain:
         self.detector_arrays = [DetectorArray(det_list, **kwargs)
                                 for det_list in opt_man.detector_setup_effects]
 
+
     def observe(self, orig_source, update=True, **kwargs):
         """
         Main controlling method for observing ``Source`` objects
@@ -150,9 +159,12 @@ class OpticalTrain:
         if update:
             self.update(**kwargs)
 
-        self.set_focus(kwargs)    # put focus back on current instrument package
+        self.set_focus(**kwargs)    # put focus back on current instrument package
 
-        source = deepcopy(orig_source)
+        # Make a copy of the Source and prepare for observation (convert to
+        # internally used units, sample to internal wavelength grid)
+        source = orig_source.make_copy()
+        source = self.prepare_source(source)
 
         # [1D - transmission curves]
         for effect in self.optics_manager.source_effects:
@@ -160,16 +172,18 @@ class OpticalTrain:
 
         # [3D - Atmospheric shifts, PSF, NCPAs, Grating shift/distortion]
         fovs = self.fov_manager.fovs
-        for fov_i, fov in enumerate(fovs):
+        for fov in fovs:
             # print("FOV", fov_i+1, "of", n_fovs, flush=True)
             # .. todo: possible bug with bg flux not using plate_scale
             #          see fov_utils.combine_imagehdu_fields
             fov.extract_from(source)
-            fov.view()
 
+            hdu_type = "cube" if self.fov_manager.is_spectroscope else "image"
+            fov.view(hdu_type)
             for effect in self.optics_manager.fov_effects:
                 fov = effect.apply_to(fov)
 
+            fov.flatten()
             self.image_planes[fov.image_plane_id].add(fov.hdu, wcs_suffix="D")
             # ..todo: finish off the multiple image plane stuff
 
@@ -180,6 +194,84 @@ class OpticalTrain:
 
         self._last_source = source
 
+
+    def prepare_source(self, source):
+        """
+        Prepare source for observation
+
+        The method is currently applied to cube fields only.
+        The source data are converted to internally used units (PHOTLAM).
+        The source data are interpolated to the waveset used by the FieldOfView
+        This is necessary when the source data are sampled on a coarser grid
+        than used internally, or if the source data are sampled on irregular
+        wavelengths.
+        For cube fields, the method assumes that the wavelengths at which the
+        cube is sampled is provided explicitely as attribute `wave` if the cube
+        ImageHDU.
+        """
+        # Convert to PHOTLAM per arcsec2
+        # ..todo: this is not sufficiently general
+
+        for cube in source.cube_fields:
+            header, data, wave = cube.header, cube.data, cube.wave
+
+            # Need to check whether BUNIT is per arcsec2 or per pixel
+            inunit = u.Unit(header['BUNIT'])
+            data = data.astype(np.float32) * inunit
+            factor = 1
+            for base, power in zip(inunit.bases, inunit.powers):
+                if (base**power).is_equivalent(u.arcsec**(-2)):
+                    conversion =(base**power).to(u.arcsec**(-2)) / base**power
+                    data *= conversion
+                    factor = u.arcsec**(-2)
+
+            data = data.to(PHOTLAM,
+                           equivalencies=u.spectral_density(wave[:, None, None]))
+
+            if factor == 1:    # Normalise to 1 arcsec2 if not a spatial density
+                # ..todo: lower needed because "DEG" is not understood, this is ugly
+                pixarea = (header['CDELT1'] * u.Unit(header['CUNIT1'].lower()) *
+                           header['CDELT2'] * u.Unit(header['CUNIT2'].lower())).to(u.arcsec**2)
+                data = data / pixarea.value    # cube is per arcsec2
+
+            data = (data * factor).value
+
+            cube.header['BUNIT'] = 'PHOTLAM/arcsec2'    # ..todo: make this more explicit?
+
+            # The imageplane_utils like to have the spatial WCS in units of "deg". Ensure
+            # that the cube is passed on accordingly
+            cube.header['CDELT1'] = header['CDELT1'] * u.Unit(header['CUNIT1'].lower()).to(u.deg)
+            cube.header['CDELT2'] = header['CDELT2'] * u.Unit(header['CUNIT2'].lower()).to(u.deg)
+            cube.header['CUNIT1'] = 'deg'
+            cube.header['CUNIT2'] = 'deg'
+
+            # Put on fov wavegrid
+            wave_min = min([fov.meta["wave_min"] for fov in self.fov_manager.fovs])
+            wave_max = max([fov.meta["wave_max"] for fov in self.fov_manager.fovs])
+            wave_unit = u.Unit(from_currsys("!SIM.spectral.wave_unit"))
+            dwave = from_currsys("!SIM.spectral.spectral_bin_width")  # Not a quantity
+            fov_waveset = np.arange(wave_min.value, wave_max.value, dwave) * wave_unit
+            fov_waveset = fov_waveset.to(u.um)
+
+            # Interpolate into new data cube.
+            # This is done layer by layer for memory reasons.
+            new_data = np.zeros((fov_waveset.shape[0], data.shape[1], data.shape[2]),
+                                dtype=np.float32)
+            for j in range(data.shape[1]):
+                cube_interp = interp1d(wave.to(u.um).value, data[:, j, :],
+                                       axis=0, kind="linear",
+                                       bounds_error=False, fill_value=0)
+                new_data[:, j, :] = cube_interp(fov_waveset.value)
+
+            cube.data = new_data
+            cube.header['CTYPE3'] = 'WAVE'
+            cube.header['CRPIX3'] = 1
+            cube.header['CRVAL3'] = wave_min.value
+            cube.header['CDELT3'] = dwave
+            cube.header['CUNIT3'] = wave_unit.name
+
+        return source
+
     def readout(self, filename=None, **kwargs):
         """
         Produces detector readouts for the observed image
@@ -202,9 +294,10 @@ class OpticalTrain:
 
         hdus = []
         for i, detector_array in enumerate(self.detector_arrays):
+            array_effects = self.optics_manager.detector_array_effects
             dtcr_effects = self.optics_manager.detector_effects
-            hdu = detector_array.readout(self.image_planes, dtcr_effects,
-                                         **kwargs)
+            hdu = detector_array.readout(self.image_planes, array_effects,
+                                         dtcr_effects, **kwargs)
 
             if filename is not None and isinstance(filename, str):
                 fname = filename
@@ -216,7 +309,7 @@ class OpticalTrain:
 
         return hdus
 
-    def set_focus(self, kwargs):
+    def set_focus(self, **kwargs):
         self.cmds.update(**kwargs)
         dy = self.cmds.default_yamls
         if len(dy) > 0 and "packages" in dy:
